#!/usr/bin/python2.5
# Copyright (C) 2008 One Laptop Per Child Association, Inc.
# Licensed under the terms of the GNU GPL v2 or later; see COPYING for details.
# Written by C. Scott Ananian <cscott@laptop.org>
"""A urlopen implementation that supports caching, seek/tell and partial
reads.

This implementation is particularly suitable for use with `zipfile.ZipFile`,
to allow you to read individual files from a remote zip file without
downloading the whole thing."""
from __future__ import division

import httplib
import os
import shutil
import tempfile
import urllib2
import urlparse
import zipfile

_cache = {}
"""Cache of temporary files used to download URLs."""

def _file_clone(f, length, mode='r'):
    """Clone the given file object to a new file object with a new file
    descriptor."""
    # disappointingly os.dup/os.dup2 returns a new file descriptor, but
    # one that shares the file pointer (offset) with the original.
    # so instead we must reopen based on the filename, which is
    # unfortunate.  UNIX needs an openfd call, like QNX has...
    cf = open(f.name, mode)
    # kludge in a length() method.
    return _HttpRangeFile(wrapped_f=cf, wrapped_length=length)

class _HttpRangeFile(object):
    """Return a file-like object which supports seek/tell via byte-range
    requests.  Will bail to an implementation which downloads the entire
    url to a temporary file and does seek/tell on it, if the target
    http server doesn't support partial GETs."""
    def __init__(self, url=None, wrapped_f=None, wrapped_length=None):
        if url is not None:
            self.f = _InnerWrapper(self, url)
            self._already_set_up = False # no connection yet.
        else:
            self._already_set_up = True
            self.f = wrapped_f
            self._length = wrapped_length
    def seek(self, offset, whence=0):
        self.f.seek(offset, whence)
    def tell(self):
        return self.f.tell()
    def read(self, size=-1):
        return self.f.read(size)
    def length(self):
        self._setup() # make sure _length is initialized
        return self._length
    def close(self):
        if self.f is not None:
            self.f.close()
            self.f = None
    def _setup(self):
        if not self._already_set_up: self.f._setup()
    def __del__(self):
        self.close()

class _InnerWrapper(object):
    """This file-like object uses byte-range requests of the remote HTTP
    server.  It bails to a full download, redirecting its outer wrapper,
    if byte-range requests fail, or if the entire file is requested."""
    def __init__(self, outer, url):
        self.outer, self.url, self.conn, self.offset = outer, url, None, 0
        self.accepts_ranges = True
    def _setup(self):
        if self.conn is not None: return # already set up
        pieces = urlparse.urlsplit(self.url)
        if pieces.scheme != 'http':
            # bail: GET to a temp file, and then open the temp file.
            return self._bail()
        self.conn = httplib.HTTPConnection(pieces.netloc)
        # do a Content-Length transaction; bail if the server doesn't play nice
        self.path = urlparse.urlunsplit \
                    ( (None, None, pieces[2], pieces[3], pieces[4]) )
        req = self.conn.request('HEAD', self.path,
                           headers={'Connection': 'Keep-Alive'})
        res = self.conn.getresponse()
        if res.status != httplib.OK:
            return self._bail()
        length = res.getheader('content-length', None)
        if length is None:
            return self._bail()
        self.length = self.outer._length = int(length)
        # verify that server accepts ranges
        if res.getheader('accept-ranges', None) != 'bytes':
            # boo.  don't bail yet, but bail on first read.
            self.accepts_ranges = False # boo
        # look at etag
        etag = res.getheader('etag', None)
        if etag is not None and etag.startswith('W/'):
            etag = None # weak etags are useless
        self.etag = etag if not None else '*'
        # read (should be zero-length)
        res.read()
        # okay, we're good to go!
        self.outer._already_set_up = True
        return
    def seek(self, offset, whence=0):
        if whence == 0:
            self.offset = offset
        elif whence == 1:
            self.offset += offset
        elif whence == 2: # offset negative in this case
            self._setup() # ensure self.length is initialized
            self.offset = self.length + offset
    def tell(self):
        return self.offset
    def read(self, size=-1):
        if size == 0: return [] # optimization
        if size < 0:
            if self.offset == 0: # optimization
                self._bail() # read whole file & cache it for the future
                return self.outer.read()
            self._setup()
            size = self.length - self.offset
        self._setup()
        assert self.offset <= self.length
        if (self.offset + size) > self.length:
            size = self.length - self.offset # fixup overlong requests
        if size == 0: return [] # we're at the end of our rope.
        if not self.accepts_ranges:
            # we actually need to do a partial read, we have to bail.
            self._bail()
            return self.outer.read(size)
        # okay, do a byte-range request.
        req = self.conn.request\
              ('GET', self.path,
               headers={'Range': 'bytes=%d-%d' % \
                        (self.offset, self.offset+size-1),
                        'If-Match': self.etag,
                        'Connection': 'Keep-Alive'})
        res = self.conn.getresponse()
        if res.status == httplib.OK:
            # hrm, server is ignoring the 'Range' request, bail!
            self._bail()
            return self.outer.read(size)
        elif res.status != httplib.PARTIAL_CONTENT:
            # some other problem, throw it.
            self.close()
            raise IOError('could not read: %s' % res.reason)
        data = res.read(size)
        assert len(data) == size
        self.offset += size
        res.close()
        return data
    def close(self):
        if self.conn is not None:
            self.conn.close()
            self.conn = None
        self.outer = None # unlink from outer wrapper
    def __del__(self):
        self.close()
    def _bail(self):
        global _cache
        # our http server doesn't support byte serving; download the full file
        # and serve from the file instead.
        if self.conn is not None:
            self.conn.close()
            self.conn = None
        tf = tempfile.NamedTemporaryFile()
        try:
            uo = urllib2.urlopen(self.url) # throws HttpError on any problem
            shutil.copyfileobj(uo, tf)
            uo.close()
        except urllib2.HTTPError:
            _cache[self.url] = None # cache the failure.
            raise
        tf.seek(0, 2) # go to the end
        self.outer._length = tf.tell()
        # record this temp file in our cache.
        _cache[self.url] = tf
        # reopen in read mode, reset to previous location
        tf = _file_clone(tf, tf.tell()) # reopen in read mode.
        tf.seek(self.offset) # reset to match where we 'were'
        self.outer.f = tf # unlink this Wrapper from its container
        self.outer._already_set_up = True

def urlopen(url, use_cache=True):
    """Return a file-like object which accesses the given URL in a way
    which supports seek/tell and partial reads.  If `use_cache` is
    True (which is the default) will store any URLs which end up being
    read completely to a persistent cache to satisfy future requests."""
    global _cache
    if use_cache and url in _cache:
        if _cache[url] is None:
            raise urllib2.HTTPError("Replaying exception")
        return _file_clone(_cache[url], _cache[url].tell())
    return _HttpRangeFile(url)

def urlretrieve(url, filename=None,
                reporthook=lambda size, ttlsize: None,
                cancelhook=lambda: False,
                use_cache=True, store_cache=True):
    """Retrieve a url to a temporary file and return the name of the
    temporary file.  Arguments are similar to `urllib.urlretrieve` but
    the return value is just the filename of the fetched file (not a
    tuple as in `urllib.urlretrieve`).  The `reporthook` gets two
    arguments: size and total size. The first invocation has size==0
    and the last one has size==total size.  The `cancelhook` should
    return True if the download is to be aborted, in which case
    urlretrieve will return None.  The `use_cache` parameter
    determines whether `urlretrieve` can/should use data already
    cached from previous calls to `urlopen` and `urlretrieve`, and the
    `store_cache` parameter determines whether downloaded files which
    were not already in the cache should be added to the cache."""
    global _cache
    BLOCK_SIZE = 64*1024
    if filename is None:
        fdst, outname = tempfile.mkstemp()
        fdst = os.fdopen(fdst, 'wb', BLOCK_SIZE)
    else:
        fdst, outname = open(filename, 'wb', BLOCK_SIZE), filename
    fsrc = urlopen(url, use_cache=use_cache)
    fcache = None if (url in _cache or not store_cache) else \
             tempfile.NamedTemporaryFile()
    i = 0
    length = fsrc.length()
    while True:
        reporthook(i*BLOCK_SIZE, length)
        if cancelhook():
            # abort!
            os.unlink(outname)
            return None
        buf = fsrc.read(BLOCK_SIZE)
        if not buf: break
        fdst.write(buf)
        if fcache is not None: fcache.write(buf)
        i+=1
    reporthook(length, length)
    if fcache is not None:
        _cache[url] = fcache
    return outname
        

def urlcleanup():
    """Clear the cache which may have been built up by previous calls to
    `urlopen` and `urlretrieve`."""
    global _cache
    _cache = {} # temp files are closed & unlinked as they become dereferenced

def _main():
    """self-test."""
    url = 'http://dev.laptop.org/~cscott/Firefox-1.xo?test#foo'
    if True:
        zf = zipfile.ZipFile(urlopen(url), 'r')
        zf.printdir()
        zf.close()
        print "----"
    import sys
    def progress(cursize, totalsize):
        sys.stdout.write('%d bytes of %d total (%3.1f%%)\r' % (cursize, totalsize, 100*(cursize / totalsize)))
        sys.stdout.flush()
    if True:
        saved_xo = urlretrieve(url, reporthook=progress, use_cache=True)
        print
        print saved_xo
    if True:
        saved_xo = urlretrieve(url, reporthook=progress)
        print
        print saved_xo
    if True:
        from sugar.bundle.activitybundle import ActivityBundle
        ab = ActivityBundle(saved_xo)
        print ab.get_name()

if __name__=='__main__': _main()
