#!/usr/bin/python2.5
# Copyright (C) 2008 One Laptop Per Child Association, Inc.
# Licensed under the terms of the GNU GPL v2 or later; see COPYING for details.
# Written by C. Scott Ananian <cscott@laptop.org>
"""Activity updater.

Checks for updates to activities and installs them."""
from __future__ import with_statement
from __future__ import division
import pygtk
pygtk.require('2.0')
import gtk, gobject
gtk.gdk.threads_init()

import gettext
import os
from threading import Thread
from gettext import gettext as _

from sugar.graphics import style

from controlpanel.sectionview import SectionView
from controlpanel.inlinealert import InlineAlert

import controlpanel.model.updater as model
from controlpanel.model.updater import _humanize_size, _svg2pixbuf

# configuration constants needed for control panel framework
CLASS = 'ActivityUpdater'
ICON = 'module-updater'
TITLE = _('Software updates')

# for debugging
_DEBUG_VIEW_ALL=False
"""View even activities with no pending updates."""

class ActivityUpdater(SectionView):
    def __init__(self, modelwrapper, alerts):
        SectionView.__init__(self)
        self.set_spacing(style.DEFAULT_SPACING)
        self.set_border_width(style.DEFAULT_SPACING * 2)
        if True: # looks better on non-sugar systems with this set to False.
            button_box = gtk.HButtonBox()
            button_box.set_layout(gtk.BUTTONBOX_END)
        else:
            button_box = gtk.HBox()
        button_box.set_spacing(style.DEFAULT_SPACING)
        hbox = gtk.HBox()
        hbox.pack_end(button_box, expand=False)
        self.size_label = gtk.Label()
        self.size_label.set_property('xalign', 0)
        self.size_label.set_justify(gtk.JUSTIFY_LEFT)
        hbox.pack_start(self.size_label, expand=True)
        self.pack_end(hbox, expand=False)
        def make_button(label_text, stock_icon):
            b = gtk.Button()
            hbox = gtk.HBox()
            hbox.set_spacing(style.DEFAULT_PADDING)
            i = gtk.Image()
            i.set_from_stock(stock_icon, gtk.ICON_SIZE_BUTTON)
            hbox.pack_start(i, expand=False)
            l = gtk.Label(label_text)
            hbox.pack_start(l, expand=False)
            b.add(hbox)
            return b
        #self.check_button = make_button(_("Check for updates"),
        #                                gtk.STOCK_REFRESH)
        self.check_button = gtk.Button(stock=gtk.STOCK_REFRESH)
        self.check_button.connect('clicked', self.refresh_cb, self)
        button_box.pack_start(self.check_button, expand=False)
        self.install_button = make_button(_("Install updates"),
                                          gtk.STOCK_OK)
        self.install_button.connect('clicked', self.download_cb, self)
        button_box.pack_start(self.install_button, expand=False)
        self.cancel_button = gtk.Button(stock=gtk.STOCK_CANCEL)
        self.cancel_button.connect('clicked', self.cancel_cb, self)
        button_box.pack_start(self.cancel_button, expand=False)
        # treeview
        self.activity_list = model.UpdateList()

        # create the TreeView using a filtered treestore
        self.ftreestore = self.activity_list.filter_new()
        if not _DEBUG_VIEW_ALL:
            self.ftreestore.set_visible_column(model.UPDATE_EXISTS)

        self.treeview = gtk.TreeView(self.ftreestore)
   
        # create some cell renderers.
        crbool = gtk.CellRendererToggle()
        crbool.set_property('activatable', True)
        crbool.set_property('xpad', style.DEFAULT_PADDING)
        def toggled_cb(crbool, path, self):
            path = self.ftreestore.convert_path_to_child_path(path)
            self.activity_list.toggle_select(path)
            self._refresh_update_size()
        crbool.connect('toggled', toggled_cb, self)

        cricon = gtk.CellRendererPixbuf()
        cricon.set_property('width', style.STANDARD_ICON_SIZE)
        cricon.set_property('height', style.STANDARD_ICON_SIZE)

        crtext = gtk.CellRendererText()
        crtext.set_property('xpad', style.DEFAULT_PADDING)
        crtext.set_property('ypad', style.DEFAULT_PADDING)
        
        # create the TreeViewColumn to display the data
        def view_func_maker(propname):
            def view_func(cell_layout, renderer, m, it):
                renderer.set_property(propname,
                                      not m.get_value(it, model.IS_HEADER))
            return view_func
        hide_func = view_func_maker('visible')
        insens_func = view_func_maker('sensitive')
        self.column_install = gtk.TreeViewColumn('Install', crbool)
        self.column_install.add_attribute(crbool, 'active', model.UPDATE_SELECTED)
        self.column_install.set_cell_data_func(crbool, hide_func)
        self.column = gtk.TreeViewColumn('Name')
        self.column.pack_start(cricon, expand=False)
        self.column.pack_start(crtext, expand=True)
        self.column.add_attribute(cricon, 'pixbuf', model.ACTIVITY_ICON)
        self.column.set_resizable(True)
        self.column.set_cell_data_func(cricon, hide_func)
        def markup_func(cell_layout, renderer, m, it):
            s = '<b>%s</b>' % m.get_value(it, model.DESCRIPTION_BIG)
            if m.get_value(it, model.IS_HEADER):
                s = '<big>%s</big>' % s
            desc = m.get_value(it, model.DESCRIPTION_SMALL)
            if desc is not None and desc != '':
                s += '\n<small>%s</small>' % desc
            renderer.set_property('markup', s)
            insens_func(cell_layout, renderer, m, it)
        self.column.set_cell_data_func(crtext, markup_func)
   
        # add tvcolumn to treeview
        self.treeview.append_column(self.column_install)
        self.treeview.append_column(self.column)

        self.treeview.set_reorderable(False)
        self.treeview.set_enable_search(False)
        self.treeview.set_headers_visible(False)
        self.treeview.set_rules_hint(True)
        self.treeview.connect('button-press-event', self.show_context_menu)

        self.top_label = gtk.Label()
        self.top_label.set_line_wrap(True)
        self.top_label.set_justify(gtk.JUSTIFY_LEFT)
        self.top_label.set_property('xalign',0)
        self.top_label.set_markup('<big>%s</big>'%_('Checking for updates...'))
        bottom_label = gtk.Label()
        bottom_label.set_line_wrap(True) # doesn't really work right =(
        bottom_label.set_justify(gtk.JUSTIFY_LEFT)
        bottom_label.set_property('xalign', 0)
        bottom_label.set_markup(_('Software updates correct errors, eliminate\nsecurity vulnerabilities, and provide new\nfeatures.'))
        vbox2 = gtk.VBox()
        vbox2.pack_start(self.top_label, expand=False)
        vbox2.pack_start(bottom_label, expand=True)
        icon = gtk.Image()
        icon.set_from_icon_name(ICON, gtk.ICON_SIZE_DIALOG)
        hbox = gtk.HBox()
        hbox.set_homogeneous(False)
        hbox.set_spacing(style.DEFAULT_SPACING)
        hbox.pack_start(icon, expand=False)
        hbox.pack_start(vbox2, expand=True)
        self.pack_start(hbox, expand=False)

        scrolled_window = gtk.ScrolledWindow()
        scrolled_window.add_with_viewport(self.treeview)
        scrolled_window.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.pack_start(scrolled_window, expand=True)
        self.activity_pane = scrolled_window
        self.progress_pane = gtk.VBox()
        self.progress_pane.set_spacing(style.DEFAULT_PADDING)
        self.progress_pane.set_border_width(style.DEFAULT_SPACING * 2)
        hbox2 = gtk.HBox()
        hbox2.set_spacing(style.DEFAULT_PADDING)
        self.progress_bar = gtk.ProgressBar()
        self.progress_label = gtk.Label()
        self.progress_label.set_property('xalign',0)
        self.progress_label.set_property('yalign',1)
        self.progress_icon = gtk.Image()
        spacer = gtk.Label()
        for widget, prop in [(self.progress_label, 'height-request'),
                             (spacer, 'height-request'),
                             (self.progress_icon, 'height-request'),
                             (self.progress_icon, 'width-request')]:
            widget.set_property(prop, style.STANDARD_ICON_SIZE)
        hbox2.pack_start(self.progress_label, expand=True)
        hbox2.pack_start(self.progress_icon, expand=False)
        self.progress_pane.pack_start(hbox2, expand=True)
        self.progress_pane.pack_start(self.progress_bar, expand=True)
        self.progress_pane.pack_start(spacer)
        self.pack_start(self.progress_pane, expand=True, fill=False)

        # show our work!
        self.show_all()
        # and start refreshing.
        self.refresh_cb(None, None)

    def download_cb(self, widget, event, data=None):
        """Invoked when the 'ok' button is clicked."""
        from sugar.bundle.activitybundle import ActivityBundle
        self.top_label.set_markup('<big>%s</big>' %
                                  _('Downloading updates...'))
        self._set_list_visibility(False)
        self._progress_cb(0, _('Starting download...'))
        self._cancel_func = lambda: self.activity_list.cancel_download()
        def progress_cb(n, extra=None, icon=None):
            gobject.idle_add(self._progress_cb, n, extra, icon)
        def do_download():
            # ensure main loop is dbus-registered
            from dbus.mainloop.glib import DBusGMainLoop
            DBusGMainLoop(set_as_default=True)
            # get activity registry
            from sugar.activity.registry import get_registry
            registry = get_registry() # requires a dbus-registered main loop
            # progress bar bookkeeping.
            counts = [0, self.activity_list.updates_selected(), 0]
            def p(n, extra, icon):
                progress_cb((n+(counts[0]/counts[1]))/2, extra, icon)
                counts[2] = n # last fraction.
            def q(n, row):
                p(n, _('Downloading %s...') % row[model.DESCRIPTION_BIG],
                  row[model.ACTIVITY_ICON])
            for f in self.activity_list.download_selected_updates(q):
                if f is None: continue # cancelled.
                ab = ActivityBundle(f)
                p(counts[2], _('Installing %s...') % ab.get_name(),
                  # XXX: should use 'get_icon_data()' here!
                  _svg2pixbuf(open(ab.get_icon()).read()))
                try:
                    if ab.is_installed():
                        ab.upgrade()
                    else:
                        # install & then select as favorite so it shows up!
                        ab.install()
                        registry.set_activity_favorite(ab.get_bundle_id(),
                                                       ab.get_activity_version(),
                                                       True)
                except:
                    pass # XXX: use alert to indicate install failure.
                os.unlink(f)
                counts[0]+=1
            # refresh when we're done.
            gobject.idle_add(self.refresh_cb, None, None, False)
        Thread(target=do_download).start()
    def cancel_cb(self, widget, event, data=None):
        """Callback when the cancel button is clicked."""
        self.cancel_button.set_sensitive(False) # no multiple clicks
        self._cancel_func()
    def refresh_cb(self, widget, event, clear_cache=True):
        """Invoked when the 'refresh' button is clicked."""
        self.top_label.set_markup('<big>%s</big>' %
                                  _('Checking for updates...'))
        self._set_list_visibility(False)
        self._cancel_func = lambda: self.activity_list.cancel_refresh()
        # unlink model from treeview, and perform actual refresh in another
        # thread.
        self.treeview.set_model(None)
        def progress_cb(n, extra=None):
            gobject.idle_add(self._progress_cb, n, extra)
        def do_refresh():
            self.activity_list.refresh(progress_cb, clear_cache=clear_cache)
            gobject.idle_add(self._refresh_done_cb)
        Thread(target=do_refresh).start()
    def _progress_cb(self, n, extra=None, icon=None):
        """Invoked in main thread during a refresh operation."""
        if n is None:
            self.progress_bar.pulse()
            self.progress_bar.set_text('')
        else:
            self.progress_bar.set_fraction(n)
            self.progress_bar.set_text('%.0f%%' % (n*100))
        self.progress_label.set_markup(extra if extra is not None else '')
        self.progress_icon.set_property('visible', icon is not None)
        if icon is not None:
            self.progress_icon.set_from_pixbuf(icon)
    def _refresh_done_cb(self):
        """Invoked in main thread when the refresh is complete."""
        self.ftreestore.refilter()
        self.treeview.set_model(self.ftreestore)
        avail = self.activity_list.updates_available()
        if avail == 0:
            header = _("Your software is up-to-date")
            self.size_label.set_markup('')
        else:
            header = gettext.ngettext("You can install %s update",
                                      "You can install %s updates", avail) \
                                      % avail
            self._refresh_update_size()
        self.top_label.set_markup('<big>%s</big>' % header)
        self._set_list_visibility(True)
    def _set_list_visibility(self, is_visible):
        """Make the activity list visible and the progress bar invisible,
        or vice-versa."""
        self.activity_pane.set_property('visible', is_visible)
        self.progress_pane.set_property('visible', (not is_visible))
        self.check_button.set_sensitive(is_visible)
        self.install_button.set_sensitive(is_visible)
        self.cancel_button.set_sensitive(not is_visible)
        self.install_button.set_property('visible', is_visible)
        self.cancel_button.set_property('visible', not is_visible)
        self.size_label.set_property('visible', is_visible)
        
    def _refresh_update_size(self):
        """Update the 'download size' label."""
        size = self.activity_list.updates_size()
        self.size_label.set_markup(_('Download size: %s')% _humanize_size(size))
        self.install_button.set_sensitive(self.activity_list.updates_selected()
                                          !=0)
        
    def show_context_menu(self, widget, event):
        """
        Show a context menu if a right click was performed on an update entry
        """
        if event.type == gtk.gdk.BUTTON_PRESS and event.button == 3:
            def cb(__, f):
                f()
                self._refresh_update_size()
            menu = gtk.Menu()
            item_select_none = gtk.MenuItem(_("_Uncheck All"))
            item_select_none.connect("activate", cb,
                                     self.activity_list.unselect_all)
            menu.add(item_select_none)
            if self.activity_list.updates_available() == 0:
                item_select_none.set_property("sensitive", False)
            item_select_all = gtk.MenuItem(_("_Check All"))
            item_select_all.connect("activate", cb,
                                    self.activity_list.select_all)
            menu.add(item_select_all)
            menu.popup(None, None, None, 0, event.time)
            menu.show_all()
            return True
        return False
        
    def delete_event(self, widget, event, data=None):
        return False # destroy me!
    def destroy(self, widget, data=None):
        gtk.main_quit()
    def main(self):
        """Start gtk main loop."""
        gtk.main()

def _main():
    """Testing code; runs updater standalone."""
    global ICON
    ICON='update-manager' # otherwise we don't find our icon stand-alone.
    window = gtk.Window(gtk.WINDOW_TOPLEVEL)
    window.set_title('Activity updater')
    window.set_size_request(425, 400)
    au = ActivityUpdater(None, None)
    window.connect('delete_event', au.delete_event)
    window.connect('destroy', au.destroy)
    window.add(au)
    au.set_border_width(style.DEFAULT_SPACING) # our window is smaller here.
    window.show()
    gtk.main()

# set PYTHONPATH to /usr/share/sugar/shell before invoking.
if __name__ == '__main__': _main ()
